#ifndef EXTERNAL_ARCHIVE_H
#define EXTERNAL_ARCHIVE_H

#include <stddef.h>
#include <stdint.h>

namespace External {
  namespace Archive {

    constexpr int MaxNameLength = 40;

    struct TarHeader
    {                              /* byte offset */
      char name[100];               /*   0 */
      char mode[8];                 /* 100 */
      char uid[8];                  /* 108 */
      char gid[8];                  /* 116 */
      char size[12];                /* 124 */
      char mtime[12];               /* 136 */
      char chksum[8];               /* 148 */
      char typeflag;                /* 156 */
      char linkname[100];           /* 157 */
      char magic[8];                /* 257 */
      char uname[32];               /* 265 */
      char gname[32];               /* 297 */
      char devmajor[8];             /* 329 */
      char devminor[8];             /* 337 */
      char padding[167];            /* 345 */
    } __attribute__((packed));

    struct File {
      const char *name;
      const uint8_t *data;
      size_t dataLength;
      bool isExecutable;
      bool readable;
    };

    bool isSane(const TarHeader* tar) ;
    bool ck_exammode(const TarHeader* tar,int index);
    //bool isExamModeAndFileNotExecutable(const TarHeader* tar) ;
    int tar_size( const TarHeader* tar);
    void tar_get_entry(const TarHeader* tar,size_t size,File & entry);
    const TarHeader* tar_nxt(const TarHeader* tar,size_t size);
  
    bool fileAtIndex(size_t index, File &entry);
    int indexFromName(const char *name);
    size_t numberOfFiles();
    size_t numberOfExecutables();
    bool executableAtIndex(size_t index, File &entry);
    uint32_t executeFile(const char *name, void * heap, const uint32_t heapSize);

  }
}

#endif
