#include "global_preferences.h"
#include <ion/exam_mode.h>
#include "flash.h"
#include <apps/external/extapp_api.h>

GlobalPreferences * GlobalPreferences::sharedGlobalPreferences() {
  static GlobalPreferences globalPreferences;
  return &globalPreferences;
}

GlobalPreferences::ExamMode GlobalPreferences::examMode() const {
  if (m_examMode == ExamMode::Unknown) {
    uint8_t mode = Ion::ExamMode::FetchExamMode();
    assert(mode >= 0 && mode < 5); // mode can be cast in ExamMode (Off, Standard, NoSym, Dutch or NoSymNoText)
    m_examMode = (ExamMode)mode;
  }
  return m_examMode;
}

GlobalPreferences::ExamMode GlobalPreferences::tempExamMode() const {
  return m_tempExamMode;
}

/* begin section (c) B. Parisse, write exam mode in *all* firmwares */
// write exam mode mode to flash at offset pos/8
void write_exammode(unsigned char * ptr,int pos,int mode,int modulo){
  int curmode=pos%modulo;
  if (curmode == mode)
    return;
  int delta=(mode+modulo-curmode)%modulo; // 0<delta<modulo, number of bits that we will set to 0
  unsigned char * target = ptr + pos/8;
  int pos8=pos % 8; // number of bits already set to 0 at target
  pos8 += delta; // 0<pos8<modulo+7
  unsigned char postab[]={0b11111111,0b1111111,0b111111,0b11111,0b1111,0b111,0b11,0b1,0}; // postab[i] is a byte with i bits set to 0 
  unsigned char tab[2];
  bool writenext=pos8>8;
  tab[0]=postab[writenext?8:pos8];
  tab[1]=postab[writenext?pos8-8:0];
#if 1 // KHICAS
  Ion::Device::Flash::WriteMemory(target, tab, writenext?2:1);
#else
  target[0]=tab[0];
  if (writenext) target[1]=tab[1];
#endif
}

// check that region is split in two parts: bits set to 0 followed by bits set to 1, returns position of the first 1 bit or -1
int pos_01(unsigned * start,int l){
  unsigned *ptr=start,* end=start+l/sizeof(unsigned);
  for (;ptr<end;++ptr){
    if (*ptr)
      break;
  }
  if (ptr==end)
    return -1;
  int pos=(ptr-start)*32;
  unsigned char * ptr8=(unsigned char *) ptr;
  if (*ptr8==0){
    pos+=8; ++ptr8;
  }
  if (*ptr8==0){
    pos+=8; ++ptr8;
  }
  if (*ptr8==0){
    pos+=8; ++ptr8;
  }
  switch (*ptr8){
  case 0b1: pos+=7; break; // 7 bits are already set to 0
  case 0b11: pos+=6; break;
  case 0b111: pos+=5; break;
  case 0b1111: pos+=4; break;
  case 0b11111: pos+=3; break;
  case 0b111111: pos+=2; break;
  case 0b1111111: pos+=1; break;
  case 0b11111111: break;
  default: return -1;
  }
  ++ptr;
  for (;ptr<end;++ptr){
    if (*ptr!=0xffffffff)
      return -1;
  }
  return pos;
}

char print_hex(int val){
  val &= 0xf;
  if (val>=0 && val<=9)
    return '0'+val;
  return 'A'+(val-10);
}

void print_hex(unsigned val,char * ptr){
  for (int i=0;i<8;++i){
    ptr[i]=print_hex(val>>(28-4*i));
  }
}

void set_exammode(int mode,int modulo){
  if (mode<0 || mode>=modulo)
    return;
  // scan external flash every dflash bytes
  unsigned * flashptr=(unsigned *) 0x90000000;
  unsigned dflash=0x10000,ntests=8*1024*1024/dflash/sizeof(unsigned);
  for (unsigned i=0;i<ntests;++i){
    unsigned * ptr=flashptr+i*dflash;
    if (0){
      extapp_pushRectUniform(0,100,320,100,0);
      char tab[9]={0};
      print_hex((unsigned)ptr,tab);
      extapp_drawTextSmall(tab,0,100,65535,0,false);
      print_hex(*(ptr),tab);
      extapp_drawTextSmall(tab,0,120,65535,0,false);
      print_hex(*(ptr+1),tab);
      extapp_drawTextSmall(tab,0,140,65535,0,false);
      print_hex(*(ptr+2),tab);
      extapp_drawTextSmall(tab,0,160,65535,0,false);
      extapp_getKey(false,0);
    }
    // kernel header?
    if (*ptr!=0xffffffff || *(ptr+1)!=0xffffffff || *(ptr+2)!= 0xdec00df0 /* f0 0d c0 de*/)
      continue;
    ptr += 0x1000/sizeof(unsigned); // exam mode buffer?
    int pos=pos_01(ptr,0x1000);
    if (pos==-1)
      continue;
    if (pos>=0x1000-8){
      Ion::Device::Flash::EraseSector(Ion::Device::Flash::SectorAtAddress((uint32_t)ptr));
      pos=0;
    }
    if (0){
      extapp_pushRectUniform(0,100,320,100,0);
      char tab[9]={0};
      print_hex((unsigned) ptr,tab);
      extapp_drawTextSmall(tab,0,100,65535,0,false);
      print_hex(pos,tab);
      extapp_drawTextSmall(tab,0,130,65535,0,false);
      extapp_getKey(false,0);
    }
    write_exammode((unsigned char *)ptr,pos,mode,modulo);
  }
}
/* end section (c) B. Parisse, */

void GlobalPreferences::setExamMode(ExamMode mode) {
  ExamMode currentMode = examMode();
  if (currentMode == mode) {
    return;
  }
  assert(mode != ExamMode::Unknown);
#if 1
  set_exammode((int)mode,4);
#else
  int8_t deltaMode = (int8_t)mode - (int8_t)currentMode;
  deltaMode = deltaMode < 0 ? deltaMode + 4 : deltaMode;
  assert(deltaMode > 0);
  Ion::ExamMode::IncrementExamMode(deltaMode);
#endif
  m_examMode = mode;
}

void GlobalPreferences::setTempExamMode(ExamMode mode) {
  m_tempExamMode = mode;
}

void GlobalPreferences::setBrightnessLevel(int brightnessLevel) {
  if (m_brightnessLevel != brightnessLevel) {
    brightnessLevel = brightnessLevel < 0 ? 0 : brightnessLevel;
    brightnessLevel = brightnessLevel > Ion::Backlight::MaxBrightness ? Ion::Backlight::MaxBrightness : brightnessLevel;
    m_brightnessLevel = brightnessLevel;
    Ion::Backlight::setBrightness(m_brightnessLevel);
  }
}
