#include <bootloader/boot.h>
#include <bootloader/slot.h>
#include <ion.h>
#include <ion/src/device/shared/drivers/reset.h>
#include <bootloader/interface.h>

#include <assert.h>

namespace Bootloader {

BootMode Boot::mode() {
  // We use the exam mode driver as storage for the boot mode
  uint8_t mode = Ion::ExamMode::FetchExamMode();

  if (mode > 3)
    return Unknown;
  
  return (BootMode) mode;
}

void Boot::setMode(BootMode mode) {
  BootMode currentMode = Boot::mode();
  if (currentMode == mode)
    return;

  assert(mode != BootMode::Unknown);
  int8_t deltaMode = (int8_t)mode - (int8_t)currentMode;
  deltaMode = deltaMode < 0 ? deltaMode + 4 : deltaMode;
  assert(deltaMode > 0);
  Ion::ExamMode::IncrementExamMode(deltaMode);
}

__attribute__((noreturn)) void Boot::boot() {
  assert(mode() != BootMode::Unknown);
  if (mode()<BootMode::SlotABootloader){
    bool ba=Slot::A().kernelHeader()->isValid(),bb=Slot::B().kernelHeader()->isValid(),bc=Slot::C().kernelHeader()->isValid();
    if (!bb && !bc) {
      // slot B and C invalid select A
      setMode(BootMode::SlotA);
    }
    else if (!ba && !bc) {
      // slot A and C invalid select B
      setMode(BootMode::SlotB);
    }
    else if (!ba && !bb) {
      // slot A and B invalid select C
      setMode(BootMode::SlotC);
    }
    if (ba && mode() == BootMode::SlotA) {
      Slot::A().boot();
    }
    if (bb && mode() == BootMode::SlotB) {
      Slot::B().boot();
    }
    if (bc && mode() == BootMode::SlotC) {
      Slot::C().boot();
    }
  }
  bootloader();
}

__attribute__ ((noreturn)) void Boot::bootloader() { 
  for(;;) {
    // Draw the interfaces and infos
    Bootloader::Interface::draw();

    // Enable USB
    Ion::USB::enable();

    // Wait for the device to be enumerated
    do {
      // If we pressed back while waiting, reset.
      uint64_t scan = Ion::Keyboard::scan();
      if (scan == Ion::Keyboard::State(Ion::Keyboard::Key::Back)) {
        Ion::Device::Reset::core();
      }
    } while (!Ion::USB::isEnumerated());

    // Launch the DFU stack, allowing to press Back to quit and reset
    Ion::USB::DFU(true);
  }
}

}
