
#include <assert.h>
#include <ion.h>

#include <bootloader/interface.h>
#include <bootloader/slot.h>
#include <bootloader/boot.h>

#include "computer.h"
#include "cable.h"

namespace Bootloader {

void Interface::drawImage(KDContext* ctx, const Image* image, int offset) {
  const uint8_t* data;
  size_t size;
  size_t pixelBufferSize;
 
  if (image != nullptr) {
    data = image->compressedPixelData();
    size = image->compressedPixelDataSize();
    pixelBufferSize = image->width() * image->height();
  } else {
    return;
  }

  KDColor pixelBuffer[4000];
  assert(pixelBufferSize <= 4000);
  assert(Ion::stackSafe()); // That's a VERY big buffer we're allocating on the stack

  Ion::decompress(
    data,
    reinterpret_cast<uint8_t *>(pixelBuffer),
    size,
    pixelBufferSize * sizeof(KDColor)
  );

  KDRect bounds((320 - image->width()) / 2, offset, image->width(), image->height());

  ctx->fillRectWithPixels(bounds, pixelBuffer, nullptr);
}

void Interface::draw() {
  KDContext * ctx = KDIonContext::sharedContext();
  ctx->fillRect(KDRect(0,0,320,240), KDColorBlack);
  drawImage(ctx, ImageStore::Computer, 109);
  drawImage(ctx, ImageStore::Cable, 172);

  ctx->drawString("Slot 1:", KDPoint(0, 0),  KDFont::SmallFont, KDColorWhite, KDColorBlack);
  ctx->drawString("Slot 2:", KDPoint(0, 13), KDFont::SmallFont, KDColorWhite, KDColorBlack);
  ctx->drawString("Slot 3:", KDPoint(0, 26), KDFont::SmallFont, KDColorWhite, KDColorBlack);
  ctx->drawString("Current:", KDPoint(0, 39),  KDFont::SmallFont, KDColorWhite, KDColorBlack);
  ctx->drawString("Beware: Numworks upgrades lock the calculator!", KDPoint(0, 52),  KDFont::SmallFont, KDColorWhite, KDColorBlack);
  ctx->drawString("! Mise a jour Numworks => calc. verrouillee!", KDPoint(0, 65),  KDFont::SmallFont, KDColorWhite, KDColorBlack);
  ctx->drawString("www-fourier.univ-grenoble-alpes.fr/~parisse/nws.html", KDPoint(0, 78),  KDFont::SmallFont, KDColorWhite, KDColorBlack);

  if (Boot::mode() == BootMode::SlotA) {
    ctx->drawString("Slot 1", KDPoint(63, 39),  KDFont::SmallFont, KDColorWhite, KDColorBlack);
  } else if (Boot::mode() == BootMode::SlotB) {
    ctx->drawString("Slot 2", KDPoint(63, 39),  KDFont::SmallFont, KDColorWhite, KDColorBlack);
  } else if (Boot::mode() == BootMode::SlotC) {
    ctx->drawString("Slot 3", KDPoint(63, 39),  KDFont::SmallFont, KDColorWhite, KDColorBlack);
  }

  Slot slots[] = {Slot::A(), Slot::B(),Slot::C()};

  for(uint8_t i = 0; i < 3; i++) {
    Slot slot = slots[i];

    if (slot.kernelHeader()->isValid() && slot.userlandHeader()->isValid()) {
      if (slot.userlandHeader()->isOmega() && slot.userlandHeader()->isUpsilon()) {
        ctx->drawString("Upsilon", KDPoint(60, i*13),  KDFont::SmallFont, KDColorWhite, KDColorBlack);
        ctx->drawString(slot.userlandHeader()->upsilonVersion(), KDPoint(147, i*13),  KDFont::SmallFont, KDColorWhite, KDColorBlack);
      } else if (slot.userlandHeader()->isOmega()) {
        ctx->drawString("Omega or Khi", KDPoint(60, i*13),  KDFont::SmallFont, KDColorWhite, KDColorBlack);
        ctx->drawString(slot.userlandHeader()->omegaVersion(), KDPoint(147, i*13),  KDFont::SmallFont, KDColorWhite, KDColorBlack);
      } else {
        ctx->drawString("Epsilon", KDPoint(60, i*13),  KDFont::SmallFont, KDColorWhite, KDColorBlack);
        ctx->drawString(slot.userlandHeader()->version(), KDPoint(147, i*13),  KDFont::SmallFont, KDColorWhite, KDColorBlack);
      }
      ctx->drawString(slot.kernelHeader()->patchLevel(), KDPoint(203, i*13),  KDFont::SmallFont, KDColorWhite, KDColorBlack);
    } else {
      ctx->drawString("Invalid", KDPoint(60, i*13),  KDFont::SmallFont, KDColorWhite, KDColorBlack);
    }
  }

}

}
